/*
  GBC OLED Regulator v1.1
  Author: Alex from insideGadgets (www.insidegadgets.com)
  Created: 19/09/2024
  Updated: 28/09/2024
  License: CC-BY-NC-SA

  Universal 8bit Graphics Library (https://github.com/olikraus/u8g2/)
  Copyright (c) 2016, olikraus@gmail.com
  All rights reserved.

  Temp/Humidity Sensor example
  Written by AA for ClosedCube
  MIT License

  Redistribution and use in source and binary forms, with or without modification, 
  are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list 
    of conditions and the following disclaimer.
    
  * Redistributions in binary form must reproduce the above copyright notice, this 
    list of conditions and the following disclaimer in the documentation and/or other 
    materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  

*/

#include <Arduino.h>
#include <EEPROM.h>
#include <string.h>
#include <U8x8lib.h>
#include <Adafruit_INA219.h>
#include "ClosedCube_HDC1080.h"
#include <LSM6.h>

#ifdef U8X8_HAVE_HW_SPI
#include <SPI.h>
#endif

#define MODE_EEPLOC 10

U8X8_SSD1306_72X40_ER_HW_I2C u8x8(/* reset=*/ U8X8_PIN_NONE);   // EastRising 0.42"
Adafruit_INA219 ina219(0x41);
ClosedCube_HDC1080 hdc1080;
LSM6 imu;

uint8_t screen[73][5];
uint8_t tiles[8] = { 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
int8_t catAnimation = 0;
uint8_t catxPos = 20;
uint8_t catAnimationCounter = 0;
uint8_t catAnimationSteps[] = {0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 8, 9, 9, 8, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, 0, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1, 0, 0};
uint8_t catEyesSteps[] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0};
uint8_t catEyesCounter = 0;
uint8_t catEyes = 0;

char blankSpace[10];
uint8_t displayMode = 0;
const int buttonPin = 5;
void setup(void) {
  u8x8.begin();
  u8x8.setPowerSave(0);
  ina219.begin();
  //ina219.setCalibration_16V_400mA();
  ina219.setCalibration_32V_1A();
  hdc1080.begin(0x40);
  imu.init();
  imu.enableDefault();
  pinMode(buttonPin, INPUT_PULLUP);
  Serial.begin(9600); 

  // Load display mode saved
  displayMode = EEPROM.read(MODE_EEPLOC);
  if (displayMode >= 7) {
    displayMode = 0;
  }
  if (displayMode == 6) {
    drawCatStart();
    catAnimationCounter = 0;
    catEyesCounter = 0;
  }
}

void loop(void) {
  if (displayMode == 0) { // Volt/Amps
    float busvoltage = 0;
    float currentmA = 0;
    float currentmAavg = 0;
    uint8_t readingmACount = 0;
    busvoltage = ina219.getBusVoltage_V();
    currentmA = ina219.getCurrent_mA();
    
    // Read current a few times for the highest values and average them
    for (uint8_t x = 0; x < 20; x++) {
      float tempmA = ina219.getCurrent_mA();
      delay(5);
      if (tempmA >= currentmA) {
        currentmA = tempmA;
        currentmAavg += tempmA;
        readingmACount++;
      }
    }

    if (readingmACount >= 2) {
      currentmAavg = currentmAavg / readingmACount;
    }
    else {
      currentmAavg = currentmA;
    }

    char voltageConvert[10];
    char voltageStr[10];
    dtostrf(busvoltage, 1, 3, voltageConvert);
    sprintf(voltageStr, "%sv  ", voltageConvert); 

    char currentConvert[10];
    char currentStr[16];
    dtostrf(currentmAavg, 2, 1, currentConvert);
    sprintf(currentStr, "%smA  ", currentConvert); 

    u8x8.setFont(u8x8_font_8x13B_1x2_r);
    u8x8.drawString(1, 0, voltageStr);
    u8x8.drawString(1, 2, currentStr);
  }
  else if (displayMode == 1) { // Volt/Amps with graph
    float busvoltage = 0;
    float currentmA = 0;
    float currentmAavg = 0;
    uint8_t readingmACount = 0;
    busvoltage = ina219.getBusVoltage_V();
    currentmA = ina219.getCurrent_mA();
    
    // Read current a few times for the highest values and average them
    for (uint8_t x = 0; x < 20; x++) {
      float tempmA = ina219.getCurrent_mA();
      delay(5);
      if (tempmA >= currentmA) {
        currentmA = tempmA;
        currentmAavg += tempmA;
        readingmACount++;
      }
    }

    if (readingmACount >= 2) {
      currentmAavg = currentmAavg / readingmACount;
    }
    else {
      currentmAavg = currentmA;
    }

    char voltageConvert[10];
    char voltageStr[10];
    dtostrf(busvoltage, 1, 3, voltageConvert);
    sprintf(voltageStr, "%sv  ", voltageConvert); 

    char currentConvert[10];
    char currentStr[16];
    dtostrf(currentmAavg, 2, 1, currentConvert);
    sprintf(currentStr, "%smA  ", currentConvert); 

    u8x8.setFont(u8x8_font_chroma48medium8_r);
    u8x8.drawString(1, 0, voltageStr);
    u8x8.drawString(1, 1, currentStr);

    uint8_t currentDot = currentmAavg / 21;
    if (currentDot > 23) {
      currentDot = 23;
    }
    if (currentDot < 8) {
      screen[71][4] = 0x80 >> currentDot;
    }
    else if (currentDot < 16) {
      screen[71][3] = 0x80 >> (currentDot - 8);
    }
    else if (currentDot < 24) {
      screen[71][2] = 0x80 >> (currentDot - 16);
    }
      
    for (uint8_t y = 2; y < 5; y++) {
      uint8_t xPixel = 0;
      for (uint8_t x = 0; x < 9; x++) {
        for (uint8_t t = 0; t < 8; t++) {
          tiles[t] = screen[xPixel][y];
          xPixel++;
        }
        u8x8.drawTile(x, y, 1, tiles);
      }
    }

    // Move display left
    for (uint8_t y = 2; y < 5; y++) {
      for (uint8_t x = 0; x < 72; x++) {
        screen[x][y] = screen[x+1][y];
      }
    }
  }
  else if (displayMode == 2) {
    // Temp/Humidity
    float temperture = hdc1080.readTemperature();
    float humidity = hdc1080.readHumidity();

    char tempConvert[10];
    char tempStr[10];
    dtostrf(temperture, 1, 2, tempConvert);
    sprintf(tempStr, "%sC  ", tempConvert); 
    
    char humidityConvert[10];
    char humidityStr[10];
    dtostrf(humidity, 1, 2, humidityConvert);
    sprintf(humidityStr, "%s%%  ", humidityConvert); 

    u8x8.setFont(u8x8_font_8x13B_1x2_r);
    u8x8.drawString(1, 0, tempStr);
    u8x8.drawString(1, 2, humidityStr);
  }
  else if (displayMode == 3) {
    // IMU Acc
    imu.read();
    char imuStrAccX[10];
    char imuStrAccY[10];
    char imuStrAccZ[10];
    
    blank_space_calc(imu.a.x); sprintf(imuStrAccX, "AX%i%s", imu.a.x, blankSpace);
    blank_space_calc(imu.a.y); sprintf(imuStrAccY, "AY%i%s", imu.a.y, blankSpace); 
    blank_space_calc(imu.a.z); sprintf(imuStrAccZ, "AZ%i%s", imu.a.z, blankSpace); 

    u8x8.setFont(u8x8_font_chroma48medium8_r);
    u8x8.drawString(0, 0, imuStrAccX);
    u8x8.drawString(0, 1, imuStrAccY);
    u8x8.drawString(0, 2, imuStrAccZ);
  }
  else if (displayMode == 4) {
    // IMU Gyro
    imu.read();
    char imuStrGyrX[10];
    char imuStrGyrY[10];
    char imuStrGyrZ[10];
    blank_space_calc(imu.g.x); sprintf(imuStrGyrX, "GX%i%s", imu.g.x, blankSpace);
    blank_space_calc(imu.g.x); sprintf(imuStrGyrY, "GY%i%s", imu.g.y, blankSpace); 
    blank_space_calc(imu.g.x); sprintf(imuStrGyrZ, "GZ%i%s", imu.g.z, blankSpace); 

    u8x8.setFont(u8x8_font_chroma48medium8_r);
    u8x8.drawString(0, 0, imuStrGyrX);
    u8x8.drawString(0, 1, imuStrGyrY);
    u8x8.drawString(0, 2, imuStrGyrZ);
  }
  else if (displayMode == 5) {
    // IMU Acc with graph
    imu.read();
    char imuStrAccX[10];
    char imuStrAccY[10];
    char imuStrAccZ[10];
    
    blank_space_calc(imu.a.x); sprintf(imuStrAccX, "AX%i%s", imu.a.x, blankSpace);
    blank_space_calc(imu.a.y); sprintf(imuStrAccY, "AY%i%s", imu.a.y, blankSpace); 
    blank_space_calc(imu.a.z); sprintf(imuStrAccZ, "AZ%i%s", imu.a.z, blankSpace); 

    u8x8.setFont(u8x8_font_chroma48medium8_r);
    u8x8.drawString(0, 0, imuStrAccX);
    u8x8.drawString(0, 1, imuStrAccY);

    int8_t xDot = imu.a.x / 3750;
    if (xDot <= -3) {
      xDot = -3;
    }
    xDot += 3;
    if (xDot > 7) {
      xDot = 7;
    }
    screen[71][2] = 0x80 >> xDot;

    int8_t yDot = imu.a.y / 3750;
    if (yDot <= -3) {
      yDot = -3;
    }
    yDot += 3;
    if (yDot > 7) {
      yDot = 7;
    }
    screen[71][3] = 0x80 >> yDot;

    int8_t zDot = imu.a.z / 3750;
    if (zDot <= -3) {
      zDot = -3;
    }
    zDot += 3;
    if (zDot > 7) {
      zDot = 7;
    }
    screen[71][4] = 0x80 >> zDot;
      
    for (uint8_t y = 2; y < 5; y++) {
      uint8_t xPixel = 0;
      for (uint8_t x = 0; x < 9; x++) {
        for (uint8_t t = 0; t < 8; t++) {
          tiles[t] = screen[xPixel][y];
          xPixel++;
        }
        u8x8.drawTile(x, y, 1, tiles);
      }
    }

    // Move display left
    for (uint8_t y = 2; y < 5; y++) {
      for (uint8_t x = 0; x < 72; x++) {
        screen[x][y] = screen[x+1][y];
      }
    }
  }
  else if (displayMode == 6) { // Cat animation
    cat_animation();
  }

  // Time
  if (displayMode == 0 || (displayMode >= 2 && displayMode <= 4)) {
    unsigned long totalSeconds = millis() / 1000;
    uint8_t hours = totalSeconds / 3600;
    uint8_t minutes = (totalSeconds % 3600) / 60;
    uint8_t seconds = (totalSeconds % 3600) % 60;
    char timeStr[16];
    sprintf(timeStr, "%dh%dm%ds  ", hours, minutes, seconds); 
    u8x8.setFont(u8x8_font_chroma48medium8_r);
    u8x8.drawString(0, 20, timeStr);
  }

  u8x8.refreshDisplay();
  delay(100);

  // Button check
  if (digitalRead(buttonPin) == LOW) {
    displayMode++;
    EEPROM.write(MODE_EEPLOC, displayMode);
    if (displayMode >= 7) {
      displayMode = 0;
    }
    while (digitalRead(buttonPin) == LOW) {
      delay(5);
    }
    delay(20);

    // Clear display
    for (uint8_t x = 0; x < 10; x++) {
      u8x8.clearLine(x);
    }
    for (uint8_t y = 0; y < 5; y++) {
      uint8_t xPixel = 0;
      for (uint8_t x = 0; x < 9; x++) {
        for (uint8_t t = 0; t < 8; t++) {
          screen[xPixel][y] = 0;
          tiles[t] = screen[xPixel][y];
          xPixel++;
        }
        u8x8.drawTile(x, y, 1, tiles);
      }
    }

    if (displayMode == 6) {
      drawCatStart();
      catAnimationCounter = 0;
      catEyesCounter = 0;
    }
  }
}

void blank_space_calc(int16_t number) {
  blankSpace[0] = 0;
  if (number >= 0) {
    if (number <= 9) {
      strncpy(blankSpace, "    ", 5);
    }
    else if (number <= 99) {
      strncpy(blankSpace, "   ", 4);
    }
    else if (number <= 999) {
      strncpy(blankSpace, "  ", 3);
    }
    else if (number <= 9999) {
     strncpy(blankSpace, " ", 2);
    }
  }
  else {
    if (number <= -999) {
      strncpy(blankSpace, " ", 3);
    }
    else if (number <= -99) {
      strncpy(blankSpace, "  ", 4);
    }
    else if (number <= -9) {
      strncpy(blankSpace, "   ", 5);
    }
  }
}

void drawCatStart(void) {
  screen[catxPos][4] = 0x0F;
  screen[catxPos+1][4] = 0x0F;
  screen[catxPos+2][4] = 0x0F;
  screen[catxPos+3][4] = 0x0F;
  screen[catxPos+0][3] = 0xFF;
  screen[catxPos+1][3] = 0xFF;
  screen[catxPos+2][3] = 0xFF;
  screen[catxPos+3][3] = 0xFF;
  screen[catxPos+0][2] = 0xF0;
  screen[catxPos+1][2] = 0xF0;
  screen[catxPos+2][2] = 0xF0;
  screen[catxPos+3][2] = 0xF0;
  screen[catxPos+4][2] = 0x0F;
  screen[catxPos+5][2] = 0x0F;
  screen[catxPos+6][2] = 0x0F;
  screen[catxPos+7][2] = 0x0F;

  // 4
  for (uint8_t x = 4; x < 12; x++) {
    screen[catxPos+x][4] = 0xF0;
  }
  for (uint8_t x = 12; x < 24; x++) {
    screen[catxPos+x][4] = 0xFF;
  }
  for (uint8_t x = 24; x < 28; x++) {
    screen[catxPos+x][4] = 0x0F;
  }
  for (uint8_t x = 28; x < 32; x++) {
    screen[catxPos+x][4] = 0xFF;
  }
  for (uint8_t x = 36; x < 40; x++) {
    screen[catxPos+x][4] = 0xFF;
  }

  // 3
  for (uint8_t x = 12; x < 16; x++) {
    screen[catxPos+x][3] = 0xF0;
  }
  for (uint8_t x = 16; x < 40; x++) {
    screen[catxPos+x][3] = 0xFF;
  }

  // 2
  for (uint8_t x = 18; x < 40; x++) {
    screen[catxPos+x][2] = 0xF0;
  }

  // 1
  for (uint8_t x = 24; x < 28; x++) {
    screen[catxPos+x][1] = 0xFF;
  }
  for (uint8_t x = 28; x < 32; x++) {
    screen[catxPos+x][1] = 0xF0;
  }
  for (uint8_t x = 32; x < 36; x++) {
    screen[catxPos+x][1] = 0xFF;
  }
  for (uint8_t x = 36; x < 40; x++) {
    screen[catxPos+x][1] = 0xF0;
  }

  // 0
  for (uint8_t x = 24; x < 28; x++) {
    screen[catxPos+x][0] = 0xFF;
  }
  for (uint8_t x = 28; x < 36; x++) {
    screen[catxPos+x][0] = 0xF0;
  }
  for (uint8_t x = 36; x < 40; x++) {
    screen[catxPos+x][0] = 0xFF;
  }
}

void clear_cat_tail(void) {
  for (uint8_t x = 1; x < 16; x++) {
    screen[catxPos-x][2] = 0x00;
    screen[catxPos-x][3] = 0x00;
    screen[catxPos-x][4] = 0x00;
  }
  for (uint8_t x = 0; x < 8; x++) {
    screen[catxPos+x][2] = 0x00;
    screen[catxPos+x][3] = 0x00;
  }
}

void cat_animation(void) {
  if (catAnimation == 0) {
    clear_cat_tail();
    for (uint8_t x = 0; x < 4; x++) {
      screen[catxPos+x][3] = 0xFF;
    }
    screen[catxPos+0][2] = 0xF0;
    screen[catxPos+1][2] = 0xF0;
    screen[catxPos+2][2] = 0xF0;
    screen[catxPos+3][2] = 0xF0;
    screen[catxPos+4][2] = 0x0F;
    screen[catxPos+5][2] = 0x0F;
    screen[catxPos+6][2] = 0x0F;
    screen[catxPos+7][2] = 0x0F;
  }
  else if (catAnimation == 1) {
    clear_cat_tail();
    for (uint8_t x = 0; x < 4; x++) {
      screen[catxPos+x][3] = 0xFF;
    }
    screen[catxPos+0][2] = 0xFF;
    screen[catxPos+1][2] = 0xFF;
    screen[catxPos+2][2] = 0xFF;
    screen[catxPos+3][2] = 0xFF;
  }
  else if (catAnimation == 2) {
    clear_cat_tail();
    for (uint8_t x = 0; x < 4; x++) {
      screen[catxPos+x][3] = 0xFF;
    }
    screen[catxPos+0][2] = 0xF0;
    screen[catxPos+1][2] = 0xF0;
    screen[catxPos+2][2] = 0xF0;
    screen[catxPos+3][2] = 0xF0;
    screen[catxPos-1][2] = 0x0F;
    screen[catxPos-2][2] = 0x0F;
    screen[catxPos-3][2] = 0x0F;
    screen[catxPos-4][2] = 0x0F;
  }
  else if (catAnimation == 3) {
    clear_cat_tail();
    for (uint8_t x = 0; x < 4; x++) {
      screen[catxPos+x][3] = 0xFF;
    }
    screen[catxPos-1][2] = 0xF0;
    screen[catxPos-2][2] = 0xF0;
    screen[catxPos-3][2] = 0xF0;
    screen[catxPos-4][2] = 0xF0;
    screen[catxPos-5][2] = 0x0F;
    screen[catxPos-6][2] = 0x0F;
    screen[catxPos-7][2] = 0x0F;
    screen[catxPos-8][2] = 0x0F;
  }
  else if (catAnimation == 4) {
    clear_cat_tail();
    for (uint8_t x = 0; x < 4; x++) {
      screen[catxPos+x][3] = 0xFF;
    }
    screen[catxPos-1][2] = 0xF0;
    screen[catxPos-2][2] = 0xF0;
    screen[catxPos-3][2] = 0xF0;
    screen[catxPos-4][2] = 0xF0;
    screen[catxPos-5][2] = 0xF0;
    screen[catxPos-6][2] = 0xF0;
    screen[catxPos-7][2] = 0xF0;
    screen[catxPos-8][2] = 0xF0;
  }
  else if (catAnimation == 5) {
    clear_cat_tail();

    screen[catxPos][3] = 0xF0;
    screen[catxPos+1][3] = 0xF0;
    screen[catxPos+2][3] = 0xF0;
    screen[catxPos+3][3] = 0xF0;
    screen[catxPos-1][3] = 0x0F;
    screen[catxPos-2][3] = 0x0F;
    screen[catxPos-3][3] = 0x0F;
    screen[catxPos-4][3] = 0x0F;
    screen[catxPos-5][3] = 0x0F;
    screen[catxPos-6][3] = 0x0F;
    screen[catxPos-7][3] = 0x0F;
    screen[catxPos-8][3] = 0x0F;
  }
  else if (catAnimation == 6) {
    clear_cat_tail();

    screen[catxPos-1][3] = 0xF0;
    screen[catxPos-2][3] = 0xF0;
    screen[catxPos-3][3] = 0xF0;
    screen[catxPos-4][3] = 0xF0;
    screen[catxPos-5][3] = 0xF0;
    screen[catxPos-6][3] = 0xF0;
    screen[catxPos-7][3] = 0xF0;
    screen[catxPos-8][3] = 0xF0;
    screen[catxPos-9][4] = 0x0F;
    screen[catxPos-10][4] = 0x0F;
    screen[catxPos-11][4] = 0x0F;
    screen[catxPos-12][4] = 0x0F;
  }
  else if (catAnimation == 7) {
    clear_cat_tail();
    
    screen[catxPos-9][4] = 0xF0;
    screen[catxPos-10][4] = 0xF0;
    screen[catxPos-11][4] = 0xF0;
    screen[catxPos-12][4] = 0xF0;
    for (uint8_t x = 0; x < 8; x++) {
      screen[catxPos-x][4] = 0x0F;
    }
  }
  else if (catAnimation == 8) {
    clear_cat_tail();
    
    screen[catxPos-9][4] = 0x0F;
    screen[catxPos-10][4] = 0x0F;
    screen[catxPos-11][4] = 0x0F;
    screen[catxPos-12][4] = 0x0F;
    for (uint8_t x = 0; x < 8; x++) {
      screen[catxPos-x][4] = 0x0F;
    }
  }
  else if (catAnimation == 9) {
    clear_cat_tail();
    
    screen[catxPos-9][3] = 0xF0;
    screen[catxPos-10][3] = 0xF0;
    screen[catxPos-11][3] = 0xF0;
    screen[catxPos-12][3] = 0xF0;
    for (uint8_t x = 0; x < 8; x++) {
      screen[catxPos-x][4] = 0x0F;
    }
  }

  // Refresh
  for (uint8_t y = 0; y < 5; y++) {
    uint8_t xPixel = 0;
    for (uint8_t x = 0; x < 9; x++) {
      for (uint8_t t = 0; t < 8; t++) {
        tiles[t] = screen[xPixel][y];
        xPixel++;
      }
      u8x8.drawTile(x, y, 1, tiles);
    }
  }
  if (catEyes == 0) {
    for (uint8_t x = 28; x < 32; x++) {
      screen[catxPos+x][1] = 0xFF;
    }
    for (uint8_t x = 36; x < 40; x++) {
      screen[catxPos+x][1] = 0xFF;
    }
  }
  else {
    for (uint8_t x = 28; x < 32; x++) {
      screen[catxPos+x][1] = 0xF0;
    }
    for (uint8_t x = 36; x < 40; x++) {
      screen[catxPos+x][1] = 0xF0;
    }
  }

  catAnimation = catAnimationSteps[catAnimationCounter];
  catAnimationCounter++;
  if (catAnimationCounter >= sizeof(catAnimationSteps)) {
    catAnimationCounter = 0;
  }
  
  catEyes = catEyesSteps[catEyesCounter];
  catEyesCounter++;
  if (catEyesCounter >= sizeof(catEyesSteps)) {
    catEyesCounter = 0;
  }
}